#include "StreamConcatFilter.h"


CStreamConcatFilter::CStreamConcatFilter()
: CMultiIOBaseFilter(NAME("Stream Concat Filter"), 0, CLS_STREAM_CONCAT)
{
	// Must call this method in sub class constructor
	Initialise();

	m_pBuffers[0] = NULL;
	m_pBuffers[1] = NULL;
	m_nSizes[0] = 0;
	m_nSizes[1] = 0;
}

CStreamConcatFilter::~CStreamConcatFilter()
{

}

CUnknown * WINAPI CStreamConcatFilter::CreateInstance( LPUNKNOWN pUnk, HRESULT *pHr )
{
	CStreamConcatFilter *pFilter = new CStreamConcatFilter();
	if (pFilter== NULL) 
	{
		*pHr = E_OUTOFMEMORY;
	}
	return pFilter;
}

HRESULT CStreamConcatFilter::Receive( IMediaSample *pSample, int nIndex )
{
	ASSERT(nIndex >= 0);
	ASSERT (nIndex < m_vInputPins.size());
	ASSERT(pSample);
	ASSERT (m_vOutputPins[0] != NULL);
	HRESULT hr;
	// Lock filter
	CAutoLock lck2(&m_csReceive);
	
	AM_SAMPLE2_PROPERTIES * const pProps = m_vInputPins[nIndex]->SampleProps();
	if (pProps->dwStreamId != AM_STREAM_MEDIA) {
		return m_vInputPins[nIndex]->Receive(pSample);
	}

	// Else copy memory
	long lSrcDataLen = pSample->GetSize();
	m_nSizes[nIndex] = lSrcDataLen;
	hr = pSample->GetPointer(&(m_pBuffers[nIndex]));
	if (FAILED(hr))
	{
		return hr;
	}

	IMediaSample * pOutSample;
	// Set up the output sample
	hr = InitializeOutputSample(pSample, &pOutSample, nIndex, 0);

	if (FAILED(hr)) {
		return hr;
	}

	// Start timing the transform (if PERF is defined)
	MSR_START(m_idTransform);

	// have the derived class transform the data

	// Get pointers to the underlying buffers.
	long lDestDataLen = pOutSample->GetSize();

	BYTE* pBufferOut;

	hr = pOutSample->GetPointer(&pBufferOut);
	if (FAILED(hr))
	{
		return hr;
	}
	
	if (m_pBuffers[0] != NULL)
	{
		//Copy this sample into the buffer
		CopyMemory(pBufferOut, m_pBuffers[0], m_nSizes[0]);
	}
	if (m_pBuffers[1] != NULL)
	{
		//Copy this sample into the buffer
		CopyMemory(pBufferOut + m_nSizes[0], m_pBuffers[1], m_nSizes[1]);
	}
	// the Transform() function can return S_FALSE to indicate that the
	// sample should not be delivered; we only deliver the sample if it's
	// really S_OK (same as NOERROR, of course.)
	if (hr == NOERROR) 
	{
		hr = m_vOutputPins[0]->GetInputPin()->Receive(pOutSample);
	} 
	else 
	{
		// S_FALSE returned from Transform is a PRIVATE agreement
		// We should return NOERROR from Receive() in this cause because returning S_FALSE
		// from Receive() means that this is the end of the stream and no more data should
		// be sent.
		if (S_FALSE == hr) 
		{

			//  Release the sample before calling notify to avoid
			//  deadlocks if the sample holds a lock on the system
			//  such as DirectDraw buffers do
			pOutSample->Release();
			return NOERROR;
		}
	}

	// release the output buffer. If the connected pin still needs it,
	// it will have addrefed it itself.
	pOutSample->Release();
	return hr;
}


int CStreamConcatFilter::InitialNumberOfInputPins()
{
	return 2;
}

int CStreamConcatFilter::InitialNumberOfOutputPins()
{
	return 1;
}

void CStreamConcatFilter::InitialiseInputTypes()
{
	AddInputType(&MEDIATYPE_Video, &MEDIASUBTYPE_RGB24, &FORMAT_VideoInfo);
}

void CStreamConcatFilter::InitialiseOutputTypes()
{
	AddOutputType(&MEDIATYPE_Video, &MEDIASUBTYPE_RGB24, &FORMAT_VideoInfo);
}


HRESULT CStreamConcatFilter::DecideBufferSize( IMemAllocator* pAlloc, ALLOCATOR_PROPERTIES* pRequestProperties, int m_nIndex )
{
	// Get media type of the output pin
	AM_MEDIA_TYPE mt;
	HRESULT hr = m_vOutputPins[m_nIndex]->ConnectionMediaType(&mt);
	if (FAILED(hr))
	{
		return hr;
	} 

	BITMAPINFOHEADER *pbmi = HEADER(mt.pbFormat);

	//RG 9/01/2008 Replacing size with correct size: The GetMediaType method has already been called at this point in time
	//For RGB formats, hence we can just use the DIBSIZE of the bitmapinfoheader
	////pRequestProperties->cbBuffer = DIBSIZE(*pbmi) * 2; 
	pRequestProperties->cbBuffer = DIBSIZE(*pbmi);

	if (pRequestProperties->cbAlign == 0)
	{
		pRequestProperties->cbAlign = 1;
	}
	if (pRequestProperties->cBuffers == 0)
	{
		pRequestProperties->cBuffers = 1;
	}
	// Release the format block.
	FreeMediaType(mt);

	// Set allocator properties.
	ALLOCATOR_PROPERTIES Actual;
	hr = pAlloc->SetProperties(pRequestProperties, &Actual);
	if (FAILED(hr)) 
	{
		return hr;
	}
	// Even when it succeeds, check the actual result.
	if (pRequestProperties->cbBuffer > Actual.cbBuffer) 
	{
		return E_FAIL;
	}
	return S_OK;
}

HRESULT CStreamConcatFilter::GetMediaType( int iPosition, CMediaType* pMediaType, int nOutputPinIndex )
{
	if (iPosition < 0)
	{
		return E_INVALIDARG;
	}
	else if (iPosition == 0)
	{
		//TODO Does pMediaType need to be freed? http://msdn2.microsoft.com/en-us/library/ms780345(VS.85).aspx
		HRESULT hr = m_vInputPins[0]->ConnectionMediaType(pMediaType);
		if (FAILED(hr))
		{
			return hr;
		}
		// If the second input is connected - modify the attributes of the output media type
		if (m_vInputPins[1]->IsConnected())
		{
			AM_MEDIA_TYPE mediaType2;
			hr = m_vInputPins[1]->ConnectionMediaType(&mediaType2);
			if (FAILED(hr))
			{
				return hr;
			}

			// Edit the bitmap properties to take the size of the end image into account
			VIDEOINFOHEADER* pVih1 = (VIDEOINFOHEADER*)pMediaType->pbFormat;
			BITMAPINFOHEADER* bmh1 = &(pVih1->bmiHeader);

			VIDEOINFOHEADER* pVih2 = (VIDEOINFOHEADER*)mediaType2.pbFormat;
			BITMAPINFOHEADER bmh2 = pVih2->bmiHeader;

			int nSizeBitmap1 = GetBitmapSize(bmh1);
			int nSizeBitmap2 = GetBitmapSize(&bmh2);

			//3) Get the Bitmapinfo header of the new image
			bmh1->biSizeImage = nSizeBitmap1 + nSizeBitmap2;
			bmh1->biHeight += bmh2.biHeight;
			//Set sample size
			pMediaType->SetSampleSize(bmh1->biSizeImage);

			//Verify that the dimensions are correct with respect to each other
			FreeMediaType(mediaType2);
		}
		return S_OK;
	}
	return VFW_S_NO_MORE_ITEMS;
}

bool CStreamConcatFilter::OnFullCreateMoreInputs()
{
	return false;
}

bool CStreamConcatFilter::OnFullCreateMoreOutputs()
{
	return false;
}